unit F_BugBrush;

(*
Notes on this program...
- the UserInterface does not use any of Delphi's UI sophistication, eg actions,
  so that you can concentrate on the effects of the code rather than on what the UI is doing
- the fonts on the panels for the UI are set to fsBold, purely for legibility when projected at BUG
- not all properties/methods/values are shown here, but they usually follow the same principles
- anything in brackets is not part of TCanvas, but a specially-written utility routine
*)


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  StdCtrls, ExtCtrls, Spin, ComCtrls, ColorGrd;


type
  TfrmBugBrush = class(TForm)
    pnlInitialStuff: TPanel;
    btnFromColour: TButton;
    shpFromColour: TShape;
    btnToColour: TButton;
    shpToColour: TShape;
    btnSolidClear: TButton;
    btnGraduClear: TButton;
    btnLine: TButton;
    spinFromX: TSpinEdit;
    spinFromY: TSpinEdit;
    spinToX: TSpinEdit;
    spinToY: TSpinEdit;
    ckbDot: TCheckBox;
    ckbArrow: TCheckBox;
    btnStatusBar: TButton;
    pnlHeader: TPanel;
    ckbSolid: TCheckBox;
    btnLines: TButton;
    btnPyramidClear: TButton;
    pnlPen: TPanel;
    pnlBrush: TPanel;
    pnlFont: TPanel;
    boxBrush: TGroupBox;
    pntBrushExample: TPaintBox;
    boxBrushColour: TGroupBox;
    cgrBrushColour: TColorGrid;
    rdoBrushStyle: TRadioGroup;
    boxPen: TGroupBox;
    pntPenExample: TPaintBox;
    boxPenColour: TGroupBox;
    cgrPenColour: TColorGrid;
    rdoPenStyle: TRadioGroup;
    boxFont: TGroupBox;
    txtFontExample: TLabel;
    boxFontColour: TGroupBox;
    cgrFontColour: TColorGrid;
    boxFontStyle: TGroupBox;
    ckbFontBold: TCheckBox;
    ckbFontItalic: TCheckBox;
    ckbFontUnderline: TCheckBox;
    ckbFontStrikeout: TCheckBox;
    pnlOnScreen: TPanel;
    pntOnScreen: TPaintBox;
    btnUndo: TButton;
    btnArmUndo: TButton;
    pnlMethods: TPanel;
    rdoMethods: TRadioGroup;
    ckbLineDot: TCheckBox;
    ckbLineArrow: TCheckBox;
    ckbLineArrowSolid: TCheckBox;
    spinRoundRectX: TSpinEdit;
    spinRoundRectY: TSpinEdit;
    btnCopy: TButton;
    btnPaste: TButton;
    btnPasteTransparent: TButton;
    btnFlip: TButton;
    btnMirror: TButton;
    mmoTextToDisplay: TMemo;
    ckbFontTransparent: TCheckBox;
    rdoTextAlignment: TRadioGroup;
    btnClear: TButton;
    boxGraduClears: TGroupBox;
    btnGraduVertical: TButton;
    btnGraduHorizontal: TButton;
    btnGraduPyramid: TButton;
    btnCut: TButton;
    boxFontNameSize: TGroupBox;
    txtFontName: TLabel;
    spinFontSize: TSpinEdit;
    Label1: TLabel;
    Label2: TLabel;
    boxPenWidth: TGroupBox;
    spinPenWidth: TSpinEdit;
    ckbBrushUseBitmap: TCheckBox;
    btnSave: TButton;
    procedure btnFromColourClick(Sender: TObject);
    procedure btnSolidClearClick(Sender: TObject);
    procedure btnToColourClick(Sender: TObject);
    procedure btnGraduClearClick(Sender: TObject);
    procedure btnLineClick(Sender: TObject);
    procedure btnStatusBarClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnLinesClick(Sender: TObject);
    procedure btnPyramidClearClick(Sender: TObject);
    procedure pntPenExamplePaint(Sender: TObject);
    procedure pntBrushExamplePaint(Sender: TObject);
    procedure cgrBrushColourMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure txtFontNameDblClick(Sender: TObject);
    procedure ckbFontTransparentClick(Sender: TObject);
    procedure cgrFontColourChange(Sender: TObject);
    procedure rdoTextAlignmentClick(Sender: TObject);
    procedure spinFontSizeChange(Sender: TObject);
    procedure ckbFontBoldClick(Sender: TObject);
    procedure pntOnScreenMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure pntOnScreenMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure pntOnScreenMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure pntOnScreenPaint(Sender: TObject);
    procedure btnUndoClick(Sender: TObject);
    procedure btnArmUndoClick(Sender: TObject);
    procedure rdoMethodsClick(Sender: TObject);
    procedure mmoTextToDisplayChange(Sender: TObject);
    procedure btnCopyClick(Sender: TObject);
    procedure btnPasteClick(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure btnFlipClick(Sender: TObject);
    procedure btnMirrorClick(Sender: TObject);
    procedure btnClearClick(Sender: TObject);
    procedure btnGraduVerticalClick(Sender: TObject);
    procedure btnGraduHorizontalClick(Sender: TObject);
    procedure btnGraduPyramidClick(Sender: TObject);
    procedure SetupTheBrush(Sender: TObject);
    procedure SetupThePen(Sender: TObject);
    procedure ckbArrowClick(Sender: TObject);
    procedure ckbLineArrowClick(Sender: TObject);
    procedure ckbBrushUseBitmapClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmBugBrush: TfrmBugBrush;

implementation

uses
  Dialogs, // for TFontDialog
  ClipBrd, // for access to the Windows clipboard
  JPEG, // if we want to save to JPG format
  BugColourControl,
  BugCommon;

{$R *.DFM}

{.$.R bug.res}
// remove the dots if you want to load the bug logo (for the header panel) from this bound-in resource

// an example of a 'cracker' class, which simple redeclares a property as public instead of private
type PanelCanvas = class(TPanel)
     public
       property Canvas;
     end;

var
  bmpLogo: TBitmap;
  TextSize: TSize;
  MarqueeRect: TRect;
  LinePoints: array of TPoint;
  bmpUndo,            // this could be a TImageList, which would allow multiple levels of undo
  bmpOffscreen,
  bmpBrush,
  bmpPaste: TBitmap;  // only assigned if actually pasting something
  bccBackground, bccForeground: TBugColourControl;

const
  // some constants to give human names to the ItemIndex values in rdoMethods
  iiPOINTER       = 0;
  iiLINE          = Succ(iiPOINTER);
  iiPOLYLINE      = Succ(iiLINE);
  iiRECTANGLE     = Succ(iiPOLYLINE);
  iiROUNDRECT     = Succ(iiRECTANGLE);
  iiELLIPSE       = Succ(iiROUNDRECT);
  iiPOLYGON       = Succ(iiELLIPSE);
  iiFLOODFILL     = Succ(iiPOLYGON);
  iiTEXTOUT       = Succ(iiFLOODFILL);
  iiMULTILINETEXT = Succ(iiTEXTOUT);
  iiLINEWITHENDS  = Succ(iiMULTILINETEXT);
  iiBUGBORDER     = Succ(iiLINEWITHENDS);
  iiCIRCLE        = Succ(iiBUGBORDER);

  // some sets to determine whether we use a Rectangle or a series of lines as the marqueeing outline
  RectangleMethods   = [iiPOINTER, iiRECTANGLE, iiROUNDRECT, iiELLIPSE, iiTEXTOUT, iiMULTILINETEXT, iiBUGBORDER];
  LineMethods        = [iiLINE, iiPOLYLINE, iiPOLYGON, iiLINEWITHENDS];
  SinglePointMethods = [iiFLOODFILL];
  TextMethods        = [iiTEXTOUT, iiMULTILINETEXT];
  CircleMethods      = [iiCIRCLE];  // we'll actually use LinePoints[0] and [1] for this

procedure TfrmBugBrush.btnFromColourClick(Sender: TObject);
begin
  shpFromColour.Brush.Color := AnotherColour;
end;

procedure TfrmBugBrush.btnToColourClick(Sender: TObject);
begin
  shpToColour.Brush.Color := AnotherColour;
end;

procedure TfrmBugBrush.btnSolidClearClick(Sender: TObject);
begin
  ClearRectangle(Self.Canvas, Self.ClientRect, shpFromColour.Brush.Color);
end;

procedure TfrmBugBrush.btnGraduClearClick(Sender: TObject);
begin
  GraduClearRectangle(Self.Canvas, Self.ClientRect, shpFromColour.Brush.Color, shpToColour.Brush.Color);
end;

procedure TfrmBugBrush.btnLineClick(Sender: TObject);
begin
  btnSolidClearClick(Sender);
  Self.Canvas.Brush.Color := shpFromColour.Brush.Color;
  Self.Canvas.Pen.Color := shpToColour.Brush.Color;
  LineWithEnds(Self.Canvas, Point(spinFromX.Value, spinFromY.Value), Point(spinToX.Value, spinToY.Value), ckbDot.Checked, ckbArrow.Checked, ckbSolid.Checked);
end;

procedure TfrmBugBrush.btnStatusBarClick(Sender: TObject);
var
  LeftRect, RightRect, BorderRect: TRect;
begin
  with pnlHeader.ClientRect
  do LeftRect := Rect(Left, Top, Left + ((Right - Left) div 2), Bottom);
  with pnlHeader.ClientRect
  do RightRect := Rect(Succ(LeftRect.Right), Top, Right, Bottom);
  with pnlHeader.ClientRect
  do BorderRect := Rect(Left + 2, Top + 2, Right - 2, Bottom - 2);
  if Assigned(bccBackground)
  then begin
         GraduClearRectangle(PanelCanvas(pnlHeader).Canvas, LeftRect, bccBackground.Colour, bccForeground.Colour, True {lefttoright});
         GraduClearRectangle(PanelCanvas(pnlHeader).Canvas, RightRect, bccForeground.Colour, bccBackground.Colour, True {lefttoright});
       end
  else begin
         GraduClearRectangle(PanelCanvas(pnlHeader).Canvas, LeftRect, shpFromColour.Brush.Color, shpToColour.Brush.Color, True {lefttoright});
         GraduClearRectangle(PanelCanvas(pnlHeader).Canvas, RightRect, shpToColour.Brush.Color, shpFromColour.Brush.Color, True {lefttoright});
       end;
  PanelCanvas(pnlHeader).Canvas.Draw(pnlHeader.Width - 85, 1, bmpLogo);
  BugBorder(PanelCanvas(pnlHeader).Canvas, BorderRect);
end;

procedure TfrmBugBrush.FormCreate(Sender: TObject);
begin
  // creation of 2 demo components in place of 2 buttons to change colours
  bccBackground:=TBugColourControl.Create(Self);
  bccBackground.Parent := pnlInitialStuff;
  bccBackground.SetBounds(85, 37, 28, 23);
  bccForeground:=TBugColourControl.Create(Self);
  bccForeground.Parent := pnlInitialStuff;
  bccForeground.SetBounds(211, 37, 28, 23);

  // set up an in-memory bitmap to hold the logo that will go in the header panel
  bmpLogo:= TBitmap.Create;
  bmpLogo.LoadFromFile('Bug_80x40.bmp'); // if you want to load a bitmap from an external BMP-file
//  bmpLogo.LoadFromResourceName(hinstance, 'bug'); // if you want to load a bitmap from a bound-in resource
  bmpLogo.Transparent:= True; // sets the Transparent colour to that of Bottom-Left pixel

  // load the example brush bitmap into an in-memory bitmap
  bmpBrush := TBitmap.Create;
  bmpBrush.LoadFromFile('Brush_8x8.bmp');

  // set up an in-memory bitmap to store the work so far, in case we want to undo something!
  bmpUndo:= TBitmap.Create; // no need to set the Size, that will be done by the Assign when bmpUndo is used

  // set up an in-memory bitmap, which will be the REAL work area, but held off-screen
  bmpOffscreen:= TBitmap.Create;
  bmpOffscreen.Width := pntOnScreen.Width;
  bmpOffscreen.Height := pntOnScreen.Height;

  // ensure that the Font controls initialise to the actual Font settings in use
  ckbFontBold.Checked := fsBold in bmpOffScreen.Canvas.Font.Style;
  ckbFontItalic.Checked := fsItalic in bmpOffScreen.Canvas.Font.Style;
  ckbFontUnderline.Checked := fsUnderline in bmpOffScreen.Canvas.Font.Style;
  ckbFontStrikeout.Checked := fsStrikeout in bmpOffScreen.Canvas.Font.Style;
  txtFontName.Caption := bmpOffScreen.Canvas.Font.Name;
  txtFontName.Transparent := ckbFontTransparent.Checked;
  spinFontSize.Value := bmpOffScreen.Canvas.Font.Size;
  txtFontExample.Font.Assign(bmpOffscreen.Canvas.Font);

  // and since we're not gonna do any real drawing on pntOnScreen, only marquee-stuff, set up the pen/brush just for that
  pntOnScreen.Canvas.Brush.Style := bsClear;
  pntOnScreen.Canvas.Pen.Style := psDot;
  pntOnScreen.Canvas.Pen.Mode := pmXOR;
end;

procedure TfrmBugBrush.FormDestroy(Sender: TObject);
begin
  bmpPaste.Free; // doesn't hurt to do this, even if it's already at nil
  bmpLogo.Free;
  bmpUndo.Free;
  bmpOffscreen.Canvas.Brush.Bitmap := nil;
  bmpOffscreen.Free;
  bmpBrush.free;
  LinePoints:= nil;  // the equivalent of free for a dynamic array
end;

procedure TfrmBugBrush.btnLinesClick(Sender: TObject);
begin
  SomeLines(Self.Canvas);
end;

procedure TfrmBugBrush.btnPyramidClearClick(Sender: TObject);
begin
//  PyramidClearRectangle(Self.Canvas, Self.ClientRect, shpFromColour.Brush.Color, shpToColour.Brush.Color);
  PyramidClearRectangle(PanelCanvas(pnlHeader).Canvas, pnlHeader.ClientRect, shpFromColour.Brush.Color, shpToColour.Brush.Color);
end;

procedure TfrmBugBrush.SetupThePen(Sender: TObject);
// NOTE Windows limits the effectiveness of Pen.Style v Pen.Width
// Width has to be 1 for any non-psSolid Style to be effective
begin
  with bmpOffscreen.Canvas.Pen
  do begin
       // uncomment this if you want to see the UI representation of the Note above
//       if (Sender = rdoPenStyle) and (rdoPenStyle.ItemIndex < 4)
//       then spinPenWidth.Value := 1
//       else if (Sender = spinPenWidth) and (spinPenWidth.Value > 1)
//            then rdoPenStyle.ItemIndex := 4;
       Color := cgrPenColour.ForegroundColor;
       case rdoPenStyle.ItemIndex of
         0: Style := psClear;
         1: Style := psDot;
         2: Style := psDash;
         3: Style := psDashDot;
         4: Style := psSolid;
       end; //case
       Width := spinPenWidth.Value;
     end;
  pntPenExample.Canvas.Pen.Assign(bmpOffscreen.Canvas.Pen);
  pntPenExamplePaint(Sender);
end;

procedure TfrmBugBrush.pntPenExamplePaint(Sender: TObject);
begin
  ClearRectangle(pntPenExample.Canvas, pntPenExample.ClientRect, boxPen.Color);
  pntPenExample.Canvas.Polyline([Point(0, pntPenExample.Height div 2),
                                 Point(pntPenExample.Width, pntPenExample.Height div 2)]);
end;

procedure TfrmBugBrush.SetupTheBrush(Sender: TObject);
// WARNING- Delphi tries to be clever...
// if you change the Style to bsClear, Delphi changes the Color to none, ie white
// if you change the Color when the Style is bsClear, Delphi changes the Style to bsSolid
// These combine to mean that you can't have a bsClear Brush with any colour to it!
begin
  // change the Style first...
  with bmpOffscreen.Canvas.Brush
  do case rdoBrushStyle.ItemIndex of
         0: Style := bsClear;
         1: Style := bsHorizontal;
         2: Style := bsVertical;
         3: Style := bsBDiagonal;
         4: Style := bsCross;
         5: Style := bsDiagCross;
         6: Style := bsSolid;
       end; //case
  //... then only change the colour if it's not going to change the Style back again!
  if rdoBrushStyle.ItemIndex <> 0 {bsClear}
  then bmpOffscreen.Canvas.Brush.Color := cgrBrushColour.BackgroundColor;

  pntBrushExample.Canvas.Brush.Assign(bmpOffscreen.Canvas.Brush);
  pntBrushExamplePaint(Sender);

  // and do the ancillary things affected by a change in the Brush colour
  txtFontExample.Color := cgrBrushColour.BackgroundColor;
  if Assigned(bmpPaste) and bmpPaste.Transparent
  then bmpPaste.TransparentColor := cgrBrushColour.BackgroundColor;
end;

procedure TfrmBugBrush.ckbBrushUseBitmapClick(Sender: TObject);
begin
  if ckbBrushUseBitmap.checked
  then begin
         bmpOffscreen.Canvas.Brush.Bitmap := bmpBrush;
         pntBrushExample.Canvas.Brush.Assign(bmpOffscreen.Canvas.Brush);
         pntBrushExamplePaint(Sender);
       end
  else begin
         bmpOffscreen.Canvas.Brush.Bitmap := nil;
         SetupTheBrush(Sender);
       end;
end; //procedure TfrmMain.ckbBrushUseBitmapClick

procedure TfrmBugBrush.pntBrushExamplePaint(Sender: TObject);
begin
  ClearRectangle(pntBrushExample.Canvas, pntBrushExample.ClientRect, boxBrush.Color);
  pntBrushExample.Canvas.Pen.Style := psClear; // avoids the border caused by the following Rectangle
  pntBrushExample.Canvas.Rectangle(pntBrushExample.ClientRect);
end;

procedure TfrmBugBrush.cgrBrushColourMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
// The Colorgrid uses Right-clicks to select backgroundcolour,
// but since this particular Colorgrid is only for backgroundcolor
// this is some code to allow Left-click to do it here (as well)
begin
  if mbLeft = Button
  then cgrBrushColour.BackgroundIndex := Y div (cgrBrushColour.Height div 16);
end;

procedure TfrmBugBrush.txtFontNameDblClick(Sender: TObject);
begin
  with TFontDialog.Create(nil)  // dynamic creation of a component that's only gonna be used very short-term
  do try
       Font.Assign(bmpOffscreen.Canvas.Font);
       if Execute
       then begin
               bmpOffscreen.Canvas.Font.Assign(Font);
               txtFontExample.Font.Assign(bmpOffscreen.Canvas.Font);
               txtFontName.Caption := Font.name;
               spinFontSize.Value := Font.Size;
               ckbFontBold.Checked := (fsBold in Font.Style);
               ckbFontItalic.Checked := (fsItalic in Font.Style);
               ckbFontUnderline.Checked := (fsUnderline in Font.Style);
               ckbFontStrikeout.Checked := (fsStrikeout in Font.Style);
               // and may need to recalculate Size of the Text extent
               mmoTextToDisplayChange(Sender);
            end; //if Execute
     finally
       Free;
     end;
end;

procedure TfrmBugBrush.ckbFontTransparentClick(Sender: TObject);
begin
  txtFontExample.Transparent := ckbFontTransparent.Checked;
end;

procedure TfrmBugBrush.cgrFontColourChange(Sender: TObject);
begin
  bmpOffscreen.Canvas.Font.Color := cgrFontColour.ForegroundColor;
  txtFontExample.Font.Assign(bmpOffscreen.Canvas.Font);
end;

procedure TfrmBugBrush.rdoTextAlignmentClick(Sender: TObject);
begin
  case rdoTextAlignment.ItemIndex of
  0: txtFontExample.Alignment := taLeftJustify;
  1: txtFontExample.Alignment := taCenter;
  2: txtFontExample.Alignment := taRightJustify;
  end;
  mmoTextToDisplayChange(Sender);
end;

procedure TfrmBugBrush.spinFontSizeChange(Sender: TObject);
begin
  bmpOffscreen.Canvas.Font.Size := spinFontSize.Value;
  txtFontExample.Font.Assign(bmpOffscreen.Canvas.Font);
  // and may need to recalculate Size of the Text extent
  mmoTextToDisplayChange(Sender);
end;

procedure TfrmBugBrush.ckbFontBoldClick(Sender: TObject);
var
  TheStyle: TFontStyles;
begin
  TheStyle:= [];
  if ckbFontBold.Checked then TheStyle := TheStyle + [fsBold];
  if ckbFontItalic.Checked then TheStyle := TheStyle + [fsItalic];
  if ckbFontUnderline.Checked then TheStyle := TheStyle + [fsUnderline];
  if ckbFontStrikeout.Checked then TheStyle := TheStyle + [fsStrikeout];
  bmpOffscreen.Canvas.Font.Style := TheStyle;
  txtFontExample.Font.Assign(bmpOffscreen.Canvas.Font);
  // and may need to recalculate size of the Text extent
  mmoTextToDisplayChange(Sender);
end;

procedure TfrmBugBrush.pntOnScreenMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft
  then begin
          if rdoMethods.ItemIndex in RectangleMethods
          then begin
                 // get rid of any existing rectangle on-screen
                 if not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
                 then pntOnScreen.Canvas.DrawFocusRect(MarqueeRect);
                 // and start off a new rectangle whose top-left is the mouse XY
                 MarqueeRect.Left := X;
                 MarqueeRect.Top := Y;
                 MarqueeRect.BottomRight := MarqueeRect.TopLeft;
               end // RectangleMethods
          else if rdoMethods.ItemIndex in LineMethods
               then begin
                      // ensure there's a starting Point
                      if Length(LinePoints) = 0
                      then begin
                            SetLength(LinePoints, 1);
                            LinePoints[0] := Point(X, Y);
                           end;
                      // and put the current Point in
                      SetLength(LinePoints, Succ(Length(LinePoints)));
                      LinePoints[high(LinePoints)] := Point(X, Y);
                    end // LineMethods
          else if rdoMethods.ItemIndex in CircleMethods
               then begin
                      // ensure there's a starting Point
                      if Length(LinePoints) <> 2
                      then SetLength(LinePoints, 2);
                      // and put the current Point in
                      LinePoints[0] := Point(X, Y);
                      LinePoints[1] := Point(X, Y);
                    end; // CircleMethods
       end; //if Button = mbLeft
end;

procedure TfrmBugBrush.pntOnScreenMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  if (rdoMethods.ItemIndex in SinglePointMethods)
  then {don't worry about mousemoves for something only needing a single point}
  else if (rdoMethods.ItemIndex = iiPOINTER) and Assigned(bmpPaste)
       then // we must be pasting
            begin
               pntOnScreen.Canvas.Draw(0, 0, bmpOffScreen);
               pntOnScreen.Canvas.Draw(X, Y, bmpPaste);
            end
  else if (rdoMethods.ItemIndex in TextMethods)
       then begin
               if not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
               then pntOnScreen.Canvas.DrawFocusRect(MarqueeRect);
               MarqueeRect.Left := X;
               MarqueeRect.Top := Y;
               MarqueeRect.Right := X + TextSize.cX;
               MarqueeRect.Bottom := Y + TextSize.cY;
               pntOnScreen.Canvas.DrawFocusRect(MarqueeRect);
            end
  else if (ssLeft in Shift)
          or
          ((rdoMethods.ItemIndex in [iiPOLYLINE, iiPOLYGON]) and (Length(LinePoints) > 1))
        then begin
               if rdoMethods.ItemIndex in LineMethods
               then begin
                      if Length(LinePoints) > 1
                      then begin
                              // remember, the pntOnScreen.Canvas.Pen is already clBlack, 1, psDot, pmXOR
                              // wipe the previous line
                              pntOnScreen.Canvas.PolyLine(Copy(LinePoints, Pred(high(LinePoints)), 2));
                              // put the new line in
                              LinePoints[high(LinePoints)] := Point(X, Y);
                              pntOnScreen.Canvas.PolyLine(Copy(LinePoints, Pred(high(LinePoints)), 2));
                           end; // there's more than 1 point
                    end // LineMethods
               else if rdoMethods.ItemIndex in RectangleMethods
                    then begin
                           if not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
                           then pntOnScreen.Canvas.DrawFocusRect(MarqueeRect);
                           // should ensure that the bottomright IS right and down from topleft, else Windows thinks it's an invalid rect
                           MarqueeRect.Right := X;
                           MarqueeRect.Bottom := Y;
                           pntOnScreen.Canvas.DrawFocusRect(MarqueeRect);
                         end // RectangleMethods
               else if rdoMethods.ItemIndex in CircleMethods
                    then begin
                           // remember, the pntOnScreen.Canvas.Pen is already clBlack, 1, psDot, pmXOR
                           if (LinePoints[0].X <> LinePoints[1].X) or (LinePoints[0].Y <> LinePoints[1].Y)
                           then // wipe the previous marquee
                                begin
                                  pntOnScreen.Canvas.PolyLine([LinePoints[0], LinePoints[1]]);
                                  Circle(pntOnScreen.Canvas, LinePoints[0], LinePoints[1]);
                                end;
                           // put the new line in
                           LinePoints[1] := Point(X, Y);
                           pntOnScreen.Canvas.PolyLine([LinePoints[0], LinePoints[1]]);
                           Circle(pntOnScreen.Canvas, LinePoints[0], LinePoints[1]);
                         end; // CircleMethods
             end; // if ssLeft or Polyline or Polygon
end;

procedure TfrmBugBrush.pntOnScreenMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  BrushWas: TBrush;
  ColourToFill: TColor;
begin
  if Button = mbLeft
  then begin
         // get rid of the latest focusrect
         if (rdoMethods.ItemIndex > iiPOINTER) and not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
         then pntOnScreen.Canvas.DrawFocusRect(MarqueeRect);
         // now decide exactly what to do in response to this mouseup
         case rdoMethods.ItemIndex of
           iiPOINTER:       if Assigned(bmpPaste)
                            then begin
                                   bmpOffScreen.Canvas.Draw(X, Y, bmpPaste);
                                   FreeAndNil(bmpPaste);
                                 end;
           iiLINE:          if Length(LinePoints) > 1
                            then begin
                                   bmpOffScreen.Canvas.Moveto(LinePoints[0].x, LinePoints[0].y);
                                   bmpOffScreen.Canvas.Lineto(LinePoints[1].x, LinePoints[1].y);
                                 end;
           iiPOLYLINE:      {don't Draw anything yet - will be done by mbRight.Up};
           iiRECTANGLE:     bmpOffScreen.Canvas.Rectangle(MarqueeRect);
           iiROUNDRECT:     bmpOffScreen.Canvas.RoundRect(MarqueeRect.Left, MarqueeRect.Top, MarqueeRect.Right, MarqueeRect.Bottom, spinRoundRectX.Value, spinRoundRectY.Value);
           iiELLIPSE:       bmpOffScreen.Canvas.Ellipse(MarqueeRect);
           iiPOLYGON:       {don't Draw anything yet - will be done by mbRight.Up};
           iiFLOODFILL:     begin
                              ColourToFill := bmpOffScreen.Canvas.Pixels[X, Y];
                              bmpOffScreen.Canvas.FloodFill(X, Y, ColourToFill, fsSurface);
                            end;
           iiTEXTOUT,
           iiMULTILINETEXT: begin
                              BrushWas := TBrush.create;
                              try //...finally
                                if ckbFontTransparent.Checked
                                then // we're gonna change the brush, so save it first
                                     begin
                                       BrushWas.Assign(bmpOffScreen.Canvas.Brush);
                                       bmpOffScreen.Canvas.Brush.Style := bsClear;
                                     end; //Transparent.Checked
                                if rdoMethods.ItemIndex = iiTEXTOUT
                                then bmpOffScreen.Canvas.TextOut(MarqueeRect.Left, MarqueeRect.Top, mmoTextToDisplay.Text)
                                else case rdoTextAlignment.ItemIndex of
                                       0 {Left}  : MultiLineTextOut(bmpOffScreen.Canvas, MarqueeRect.Left, MarqueeRect.Top, mmoTextToDisplay.Text, taLeftJustify);
                                       1 {Centre}: MultiLineTextOut(bmpOffScreen.Canvas, MarqueeRect.Left, MarqueeRect.Top, mmoTextToDisplay.Text, taCenter);
                                       2 {Right} : MultiLineTextOut(bmpOffScreen.Canvas, MarqueeRect.Left, MarqueeRect.Top, mmoTextToDisplay.Text, taRightJustify);
                                     end; //if, case
                                if ckbFontTransparent.Checked
                                then bmpOffScreen.Canvas.Brush.Assign(BrushWas);
                              finally
                                BrushWas.Free;
                              end; //t..f
                            end;
           iiLINEWITHENDS:  if Length(LinePoints) > 1
                            then LineWithEnds(bmpOffScreen.Canvas, LinePoints[0], LinePoints[1], ckbLineDot.Checked, ckbLineArrow.Checked, ckbLineArrowSolid.Checked);
           iiBUGBORDER:     BugBorder(bmpOffScreen.Canvas, MarqueeRect);
           iiCIRCLE:        Circle(bmpOffScreen.Canvas, LinePoints[0], LinePoints[1]);
         end; //case
         if rdoMethods.ItemIndex > iiPOINTER
         then begin
                 // we've probably done something which now needs to be shown on-screen (except the polyline/polygon)
                 if not (rdoMethods.ItemIndex in [iiPOLYLINE, iiPOLYGON])
                 then pntOnScreen.Repaint;
                 // we've probably done something so clear the marquee definition
                 if rdoMethods.ItemIndex in RectangleMethods
                 then MarqueeRect:= Rect(0,0,0,0)
                 else if rdoMethods.ItemIndex in [iiLINE, iiLINEWITHENDS, iiCIRCLE]
                      then SetLength(LinePoints, 0);
              end; // if not the pointer
       end // mbLeft
  else if (Button = mbRight) and (rdoMethods.ItemIndex in [iiPOLYLINE, iiPOLYGON])
       then // mbRight terminates a polyline/polygon
            begin
              case rdoMethods.ItemIndex of
                iiPOLYLINE:  bmpOffScreen.Canvas.Polyline(LinePoints);
                iiPOLYGON:   bmpOffScreen.Canvas.Polygon(LinePoints);
              end; //case
              pntOnScreen.Repaint;
              // clear the marqueerect
              SetLength(LinePoints, 0);
            end; // mbRight polyline/polygon
end;

procedure TfrmBugBrush.pntOnScreenPaint(Sender: TObject);
// this is the Paint routine called by Windows any time it needs to redraw pntOnScreen
begin
  pntOnScreen.Canvas.Draw(0, 0, bmpOffScreen);
end;

procedure TfrmBugBrush.btnUndoClick(Sender: TObject);
begin
  btnUndo.Enabled := False;
  if not bmpUndo.Empty
  then begin
         bmpOffScreen.Assign(bmpUndo);
         pntOnScreen.Repaint;
       end;
  if Assigned(bmpPaste)
  then FreeAndNil(bmpPaste);
end;

procedure TfrmBugBrush.btnArmUndoClick(Sender: TObject);
begin
  bmpUndo.Assign(bmpOffScreen);
  btnUndo.Enabled := True;
end;

procedure TfrmBugBrush.rdoMethodsClick(Sender: TObject);
begin
  if not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
  then begin
         pntOnScreen.Canvas.DrawFocusRect(MarqueeRect);
         MarqueeRect := Rect(0,0,0,0);
       end;
  // any change in the tool (method) forces a save of the work so far so an undo can be done
  btnArmUndoClick(Sender);
  if (rdoMethods.ItemIndex in TextMethods)
  then begin
         if mmoTextToDisplay.Text = ''
         then // there's no text to display - woopsie!
              rdoMethods.ItemIndex := iiPOINTER // not very user-friendly, but adequate safety for this demo
         else // calculate the size of the text to display
              mmoTextToDisplayChange(Sender);
       end;
end;

procedure TfrmBugBrush.mmoTextToDisplayChange(Sender: TObject);
begin
  if rdoMethods.ItemIndex = iiTEXTOUT
  then TextSize := bmpOffScreen.Canvas.TextExtent(mmoTextToDisplay.Text)
  else TextSize := MultiLineTextExtent(bmpOffScreen.Canvas, mmoTextToDisplay.Text);
end;

procedure TfrmBugBrush.btnCopyClick(Sender: TObject);
var
  bmpCopy: TBitmap;
begin
  if (rdoMethods.ItemIndex = iiPOINTER)
     and not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
  then begin
         bmpCopy:= TBitmap.Create;
         try {..finally}
           bmpCopy.Width := MarqueeRect.Right - MarqueeRect.Left;
           bmpCopy.Height := MarqueeRect.Bottom - MarqueeRect.Top;
           bmpCopy.Canvas.CopyRect(Rect(0,0,bmpCopy.Width, bmpCopy.Height), bmpOffScreen.Canvas, MarqueeRect);
           Clipboard.Assign(bmpCopy);
         finally
           bmpCopy.Free;
         end; {try..finally}
         if Sender = btnCut
         then begin
                btnArmUndoClick(Sender);
                ClearRectangle(bmpOffscreen.Canvas, MarqueeRect, cgrBrushColour.BackgroundColor);
                pntOnScreen.Repaint;
              end;
       end;
end;

procedure TfrmBugBrush.btnPasteClick(Sender: TObject);
begin
  FreeAndNil(bmpPaste);
  if Clipboard.HasFormat(CF_BITMAP)
  then begin
         bmpPaste := TBitmap.Create;
         bmpPaste.Assign(Clipboard);
         bmpPaste.Transparent:= (Sender = btnPasteTransparent);
         if bmpPaste.Transparent
         then bmpPaste.TransparentColor := cgrBrushColour.BackgroundColor;
         rdoMethods.ItemIndex := iiPOINTER;
       end
  else Beep;
end;

procedure TfrmBugBrush.FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if Shift = [ssCtrl]
  then case Key of
         Ord('C'): // ctrl-C = Copy
                   btnCopyClick(btnCopy);
         Ord('X'): // ctrl-X = Cut
                   btnCopyClick(btnCut);
         Ord('V'): // ctrl-V = Paste
                   btnPasteClick(btnPaste);
         Ord('E'): // ctrl-E = PasteTransparent
                   btnPasteClick(btnPasteTransparent);
         Ord('F'): // ctrl-F = Flip whatever is being pasted
                   btnFlipClick(btnFlip);
         Ord('M'): // ctrl-M = Mirror whatever is being pasted
                   btnMirrorClick(btnMirror);
         Ord('S'): // ctrl-S = Save the current bmp, either as BMP or as JPG
                   btnSaveClick(btnSave);
         Ord('Z'): // ctrl-Z = Undo
                   btnUndoClick(btnUndo);
       end; //case
end;

procedure TfrmBugBrush.btnFlipClick(Sender: TObject);
begin
  if (rdoMethods.ItemIndex = iiPOINTER) and Assigned(bmpPaste)
  then FlipBitmap(bmpPaste);
end;

procedure TfrmBugBrush.btnMirrorClick(Sender: TObject);
begin
  if (rdoMethods.ItemIndex = iiPOINTER) and Assigned(bmpPaste)
  then MirrorBitmap(bmpPaste);
end;

procedure TfrmBugBrush.btnClearClick(Sender: TObject);
begin
  btnArmUndoClick(Sender);
  ClearRectangle(bmpOffscreen.Canvas, pntOnscreen.ClientRect, cgrBrushColour.BackgroundColor);
  pntOnScreen.Repaint;
  if not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
  then MarqueeRect := Rect(0,0,0,0);
end;

procedure TfrmBugBrush.btnGraduVerticalClick(Sender: TObject);
begin
  btnArmUndoClick(Sender);
  GraduClearRectangle(bmpOffscreen.Canvas, pntOnscreen.ClientRect, cgrBrushColour.BackgroundColor, cgrPenColour.ForegroundColor);
  pntOnScreen.Repaint;
  if not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
  then MarqueeRect := Rect(0,0,0,0);
end;

procedure TfrmBugBrush.btnGraduHorizontalClick(Sender: TObject);
begin
  btnArmUndoClick(Sender);
  GraduClearRectangle(bmpOffscreen.Canvas, pntOnscreen.ClientRect, cgrBrushColour.BackgroundColor, cgrPenColour.ForegroundColor, True);
  pntOnScreen.Repaint;
  if not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
  then MarqueeRect := Rect(0,0,0,0);
end;

procedure TfrmBugBrush.btnGraduPyramidClick(Sender: TObject);
begin
  btnArmUndoClick(Sender);
  PyramidClearRectangle(bmpOffscreen.Canvas, pntOnscreen.ClientRect, cgrBrushColour.BackgroundColor, cgrPenColour.ForegroundColor);
  pntOnScreen.Repaint;
  if not IsRectEmpty(MarqueeRect)   // Windows API, but simple to port to Kylix
  then MarqueeRect := Rect(0,0,0,0);
end;

procedure TfrmBugBrush.ckbArrowClick(Sender: TObject);
begin
  ckbSolid.Enabled := ckbArrow.Checked;
end;

procedure TfrmBugBrush.ckbLineArrowClick(Sender: TObject);
begin
  ckbLineArrowSolid.Enabled := ckbLineArrow.Checked;
end;

procedure TfrmBugBrush.btnSaveClick(Sender: TObject);
var
  ucExt: string;
begin
  with TSaveDialog.Create(nil)
  do try
       DefaultExt := 'JPG';
       Filter := 'JPEGs, Bitmaps|*.jpg;*.bmp';
       Title := 'Save BUG work as...';
       if Execute
       then begin
              ucExt:= UpperCase(ExtractFileExt(FileName));
              if ucExt = '.BMP'
              then bmpOffScreen.SaveToFile(FileName)
              else if ucExt = '.JPG'
                   then with TJPEGImage.Create
                        do try
                             Assign(bmpOffScreen);
                             SaveToFile(FileName);
                           finally
                             Free;
                           end; // with TJPEGImage
            end; // if Execute
     finally
       Free;
     end; // with TSaveDialog
end;

end.


